function _SglViewerHandler(meshURL) {
	this.meshURL = meshURL || null;
}

_SglViewerHandler.prototype = {
	loadMesh : function (url) {
		this.meshGL       = null;
		this.meshJS       = new SglMeshJS();
		if (this.meshGL) {
			this.meshGL.destroy();
		}
		this.meshGL  = null;
		this.meshJS  = new SglMeshJS();
		this.meshURL = url;

		var that = this;

		this.meshJS.importJSON(url, true, function(m, url) {
			if (!m.isValid) return;
			var gl = that.ui.gl;
			that.meshGL = that.meshJS.toPackedMeshGL(gl, "triangles", 65000);
			var bbox = that.meshJS.calculateBoundingBox("position");
			var bc = bbox.center;
			var s  = 1.0 / bbox.diagonal;
			that.centerMatrix = sglMulM4(sglScalingM4C(s, s, s), sglTranslationM4C(-bc[0], -bc[1], -bc[2]));
			that.trackball.reset();
			that.ui.requestDraw();
		});
	},

	load : function (gl) {
		var vshd = "";
		vshd += "#ifdef GL_ES\n";
		vshd += "precision highp float;\n";
		vshd += "#endif\n";
		vshd += "uniform   mat4 u_mvp;\n";
		vshd += "uniform   mat3 u_normal_mat;\n";
		vshd += "attribute vec4 a_position;\n";
		vshd += "attribute vec3 a_normal;\n";
		vshd += "attribute vec3 a_color;\n";
		vshd += "varying   vec3 v_normal;\n";
		vshd += "varying   vec3 v_color;\n";
		vshd += "void main(void)\n";
		vshd += "{\n";
		vshd += "	v_normal    = u_normal_mat * a_normal;\n";
		vshd += "	v_color     = a_color;\n";
		vshd += "	gl_Position = u_mvp * a_position;\n";
		vshd += "}\n";

		var fshd = "";
		fshd += "#ifdef GL_ES\n";
		fshd += "precision highp float;\n";
		fshd += "#endif\n";
		fshd += "varying vec3 v_normal;\n";
		fshd += "varying vec3 v_color;\n";
		fshd += "void main(void)\n";
		fshd += "{\n";
		fshd += "	const vec3 c_light_dir = vec3(0.0, 0.0, -1.0);\n";
		fshd += "	vec3  normal  = normalize(v_normal);\n";
		fshd += "	float lambert = max(dot(normal, -c_light_dir), 0.0);\n";
		fshd += "	vec3  color   = v_color * lambert;\n";
		fshd += "	gl_FragColor = vec4(color, 1.0);\n";
		fshd += "}\n";

		this.prog         = new SglProgram(gl, [vshd], [fshd]);
		this.xform        = new SglTransformStack();
		this.viewMatrix   = sglLookAtM4C(0.0, 0.0, 2.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
		this.trackball    = new SglTrackball();
		this.colorIndex   = this.prog.attributes["a_color"].index;
		this.normalIndex  = this.prog.attributes["a_normal"].index;
		this.centerMatrix = sglIdentityM4();;
		this.meshGL       = null;
		this.meshJS       = null;

		if (this.meshURL) {
			this.loadMesh(this.meshURL);
		}
	},

	unload : function (gl) {
	},

	keyDown : function (gl, keyCode, keyString) {
	},

	keyUp : function (gl, keyCode, keyString) {
	},

	keyPress : function (gl, keyCode, keyString) {
		if (keyString == "R") {
			this.trackball.reset();
		}
	},

	mouseDown : function (gl, button, x, y) {
	},

	mouseUp : function (gl, button, x, y) {
	},

	mouseMove : function (gl, x, y) {
		var ui = this.ui;
		var ax1 = (x / (ui.width  - 1)) * 2.0 - 1.0;
		var ay1 = (y / (ui.height - 1)) * 2.0 - 1.0;
		var action = SGL_TRACKBALL_NO_ACTION;
		if ((ui.mouseButtonsDown[0] && ui.keysDown[17]) || ui.mouseButtonsDown[1]) {
			action = SGL_TRACKBALL_PAN;
		}
		else if (ui.mouseButtonsDown[0]) {
			action = SGL_TRACKBALL_ROTATE;
		}
		this.trackball.action = action;
		this.trackball.track(this.viewMatrix, ax1, ay1, 0.0);
		this.trackball.action = SGL_TRACKBALL_NO_ACTION;
	},

	mouseWheel: function (gl, wheelDelta, x, y) {
		var action = (this.ui.keysDown[16]) ? (SGL_TRACKBALL_DOLLY) : (SGL_TRACKBALL_SCALE);
		var factor = (action == SGL_TRACKBALL_DOLLY) ? (wheelDelta * 0.3) : ((wheelDelta < 0.0) ? (1.10) : (0.90));
		this.trackball.action = action;
		this.trackball.track(this.viewMatrix, 0.0, 0.0, factor);
		this.trackball.action = SGL_TRACKBALL_NO_ACTION;
	},

	click : function (gl, button, x, y) {
	},

	dblClick : function (gl, button, x, y) {
	},

	resize : function (gl, width, height) {
	},

	update : function (gl, dt) {
	},

	draw : function (gl) {
		var w = this.ui.width;
		var h = this.ui.height;

		gl.clearColor(1.0, 1.0, 1.0, 1.0);
		gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT | gl.STENCIL_BUFFER_BIT);

		gl.viewport(0, 0, w, h);

		this.xform.projection.loadIdentity();
		this.xform.projection.perspective(sglDegToRad(45.0), w/h, 0.1, 10.0);

		this.xform.view.load(this.viewMatrix);

		this.xform.model.load(this.trackball.matrix);
		this.xform.model.multiply(this.centerMatrix);

		gl.enable(gl.DEPTH_TEST);
		gl.enable(gl.CULL_FACE);

		if (this.meshGL) {
			if (!this.meshGL.vertices.attributes["color"]) {
				gl.vertexAttrib3f(this.colorIndex, 0.8, 0.8, 0.8);
			}
			if (!this.meshGL.vertices.attributes["normal"]) {
				gl.vertexAttrib3f(this.normalIndex, 0.0, 0.0, 1.0);
			}
			var uniforms = {
				u_mvp        : this.xform.modelViewProjectionMatrix,
				u_normal_mat : this.xform.viewSpaceNormalMatrix
			};
			sglRenderMeshGLPrimitives(this.meshGL, "triangles", this.prog, null, uniforms);
		}

		gl.disable(gl.DEPTH_TEST);
		gl.disable(gl.CULL_FACE);
	}
};

function SglViewer(canvasID, meshURL) {
	this._viewer = new _SglViewerHandler(meshURL);
	sglRegisterCanvas(canvasID, this._viewer, 0);
};

SglViewer.prototype = {
	get meshURL()    { return this._viewer.meshURL; },
	set meshURL(url) { this._viewer.loadMesh(url);  }
};

function sglCreateMeshViewer(canvasID, meshURL) {
	return new SglViewer(canvasID, meshURL);
}
